// Functions for image processing

#include <opencv/cv.h>
#include <opencv/highgui.h>
#include <stdio.h>
#include <stdlib.h>

#include "mesaRangeData.h"
#include "myutil.h"
#include "ip.h"
#include "pixel.h"

//---------------------------------------------------------------------------------------------//
/*
 * Name		balloc

 * Description	Two dimension memory allocator
 * Compilation	cc -c balloc.c
 * Author	Sergio Antoy
 * Date		21 January 1985
 *
 */

/*
 * It is allocated a pointer to an extra row, but not the row.
 * It is assumed that "calloc" sets to NULL the value of the
 * extra pointer and that "balloc" sets to not NULL the value
 * of all the other pointers to rows.
 * In this way "bfree" can know the number of rows without any
 * argument explicitly passed.
 *
 */

char **balloc(int nrows, int ncols, int elem_size)
{

  int r;
  char **array;

  if (odd(nrows)) nrows++;
  if (odd(ncols)) ncols++;

  if ((array = (char **)calloc((unsigned)(nrows + 1), 
                               (unsigned)sizeof(char *))) == NULL)
    return(NULL);

  for (r = 0 ; r < nrows ; r++)				/* allocate each row */
    {
      if ((array[r] = (char *)calloc((unsigned)ncols, elem_size)) == NULL)
	return(NULL);
    }

  return(array);					/* return pointer */
}

void bfree(char **array)
{

  int r;

  r = 0;
  while (array[r] != NULL) {
    free(array[r]);
    r++; }
  free((char *)array);

}

//---------------------------------------------------------------------------------------------//
// labeling

#define MAXTABLE	10000

void check_label_table(int *table, int maxk, int labelold, int labelnew)
{
   int k;

   for (k = maxk; k > 0; k--)
     if (table[k] == labelold) table[k] = labelnew;
}

int labeling(unsigned int **img, int width, int height) 
{
  int     i, j, k, kk, valuel, valueu;
  int     table[MAXTABLE], table2[MAXTABLE];
  
  /* initialize table */
  for (k = 0; k < MAXTABLE; k++) table[k] = k; 
  
  k = 0;
  /* Pڂ̑ */
  for(i = 1; i < height-1; i++){
      for(j = 1; j < width-1; j++)
		if (img[i][j] == FMAX){
			if (img[i-1][j] == FMIN){
				if (img[i][j-1] == FMIN){
					k = k + 1;
					img[i][j] = k;
				}
				else{
					img[i][j] = img[i][j-1];
				}
			}
			else{    /* img[i-1][j] == FMAX */
				if (img[i][j-1] == FMIN){
					img[i][j] = img[i-1][j];
				}
				else{
					valuel = img[i][j-1];
					valueu = img[i-1][j];
					if (table[valuel] == table[valueu])
						img[i][j] = valuel;
					else if (table[valuel] > table[valueu]){
						img[i][j] = valueu;
						check_label_table(table, k, table[valuel], table[valueu]);
					}
					else{
						img[i][j] = valuel;
						check_label_table(table, k, table[valueu], table[valuel]);
					}
				}
			}
		}
	}

  /* check label table */
  kk = 1;
  for (i = 1; i <= k; i++){
	  for (j = 0; j < kk; j++)
		  if (table[i] == table2[j]) break;
	  if (j < kk) table[i] = j;
      else{
		table2[kk] = table[i];
		table[i] = kk;
		kk++;
	  }
  }

  /* Qڂ̑ */
  for(i = 1; i < height-1; i++){
	  for(j = 1; j < width-1; j++)
		  if (img[i][j] > FMIN)
			  img[i][j] = table[img[i][j]];
  }

  return(kk-1);
}

// find largest region
int find_largest_region(int region_num, unsigned int **img, int width, int height)
{
	int *sizes, maxsize, max_r;

	sizes = CNew(int, region_num+1);

	for (int r = 1; r <= region_num; r++) sizes[r] = 0;

	for (int h = 0; h < height; h++){
		for (int w = 0; w < width; w++){
			if (img[h][w] > FMIN){
				fprintf(stderr, "img[%d][%d] = %d\n", h, w, img[h][w]);
				sizes[img[h][w]]++;
			}
		}
	}

	maxsize = 0;
	for (int r = 1; r <= region_num; r++){
		printf("%d size : %d\n",r, sizes[r]);
		if (sizes[r] > maxsize){
			maxsize = sizes[r];
			max_r = r;
		}
	}

	return max_r;
}

//-----------------------------------------------------------------------------------------------//
//
//  OpenCV pixel manipulation functions:
//
//-----------------------------------------------------------------------------------------------//

unsigned char GetIred(IplImage *image,int x,int y){

  if(image->nChannels!=3){
    fprintf(stderr,"Error GetIred\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(GetIred):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  return ((unsigned char*)(image->imageData + image->widthStep*y))[x*3+2];

}

unsigned char GetIgreen(IplImage *image,int x,int y){
  if(image->nChannels!=3){
    fprintf(stderr,"Error GetIgreen\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(GetIgreen):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  return ((unsigned char*)(image->imageData + image->widthStep*y))[x*3+1];

}

unsigned char GetIblue(IplImage *image,int x,int y){
  if(image->nChannels!=3){
    fprintf(stderr,"Error GetIblue\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(GetIblue):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  return ((unsigned char*)(image->imageData + image->widthStep*y))[x*3+0];

}

void SetIred(IplImage *image,int x,int y,unsigned char value){
  if(image->nChannels!=3){
    fprintf(stderr,"Error SetIred\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(SetIred):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  ((unsigned char*)(image->imageData + image->widthStep*y))[x*3+2]=value;
}

void SetIgreen(IplImage *image,int x,int y,unsigned char value){

  if(image->nChannels!=3){
    fprintf(stderr,"Error SetIgreen\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(SetIgreen):x=%d,y=%d\n",x,y);
    exit(-1);
  }


  ((unsigned char*)(image->imageData + image->widthStep*y))[x*3+1]=value;
}

void SetIblue(IplImage *image,int x,int y,unsigned char value){
  if(image->nChannels!=3){
    fprintf(stderr,"Error SetIblue\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(SetIblue):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  ((unsigned char*)(image->imageData + image->widthStep*y))[x*3+0]=value;
}


unsigned char GetIgray(IplImage *image,int x,int y){
  if(image->nChannels!=1){
    fprintf(stderr,"Error GetIgray\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(GetIgray):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  return ((unsigned char*)(image->imageData + image->widthStep*y))[x];

}

void SetIgray(IplImage *image,int x,int y,unsigned char value){
  if(image->nChannels!=1){
    fprintf(stderr,"Error SetIgray\n");
    exit(-1);
  }

  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(SetIgray):x=%d,y=%d\n",x,y);
    exit(-1);
  }

  ((unsigned char*)(image->imageData + image->widthStep*y))[x]=value;
}


float GetI32f(IplImage *image,int x,int y){
  
  if(image->nChannels!=1){
    fprintf(stderr,"GetI32f\n");
    exit(-1);
  }
  
  if(x<0 || x>=image->width || y<0 || y>=image->height){
    fprintf(stderr,"index error(GetI32f):x=%d,y=%d\n",x,y);
    exit(-1);
  }


  float *p=(float *)image->imageData;
  return p[x+y*image->widthStep/sizeof(float)];
}
